#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff


class SendResetEmail(ff.DomainService):
    _subdomain: str = None
    _reset_url: str = None
    _from_address: str = None

    def __call__(self, username: str, cache_id: str, **kwargs):
        reset_url = f'{self._reset_url}?request_id={cache_id}'
        html_body = self._gen_html_body(reset_url)
        text_body = self._gen_text_body(reset_url)

        data = {
            'subject': 'PWR Lab Password Reset',
            'text_body': text_body,
            'html_body': html_body,
            'from_address': self._from_address,
            'to_address': [username],
            'cc_addresses': [],
            'bcc_addresses': []
        }
        try:
            x = self.invoke('messaging.SendSESEmail', data)
            return True
        except Exception as e:
            return False

    def _gen_html_body(self, reset_url):
        return f"""<h1 style="text-align: center;">Password Reset</h1>
<br><br>
Trouble signing in? We received a request to reset your password.
<br><br>
If you would like to continue with resetting the password for your account, <a href="{reset_url}" >Click Here</a>. Or copy and paste this URL into your browser: <br><a href="{reset_url}" >{reset_url}</a>
<br><br>
If you did not request this password reset, please ignore this message.
<br><br>
Cheers,
<br>
-PwrLab Team
"""

    def _gen_text_body(self, reset_url):
        return f"""Password Reset

Trouble signing in? We received a request to reset your password.

If you would like to continue with resetting the password for your account, copy and paste this URL into your browser: {reset_url}

If you did not request this password reset, please ignore this message.

Cheers,
-PwrLab Team
"""



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain


@ff.authenticator()
class OAuthAuthenticator(ff.Handler, ff.LoggerAware):
    _oauth_authenticator: domain.OAuthAuthenticator = None

    def handle(self, message: ff.Message, **kwargs):
        return self._oauth_authenticator(message, **kwargs)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
from .generic_oauth_middleware import GenericOauthDomainMiddleware


class OAuthAuthorizeRequest(GenericOauthDomainMiddleware):

    def __call__(self, message: ff.Message, **kwargs):
        token = None
        message = self._fix_email(message)
        try:
            if not message.access_token:
                token = self._get_token()
                if not token:
                    return False
                message.access_token = token
            else:
                token = message.access_token
        except AttributeError:
            token = self._get_token()
            if not token:
                return False
            message.access_token = token
        if not message.access_token and not token:
            return False
        if message.access_token.lower().startswith('bearer'):
            message.access_token = message.access_token.split(' ')[-1]
        try:
            self._decode_token(token, self._kernel.user.id)
        except:
            raise ff.UnauthorizedError()

        message.token = message.access_token
        validated, resp = self._oauth_provider.verify_request(message, message.scopes)

        return validated

    def _get_token(self):
        token = None
        try:
            token = self._retrieve_token_from_http_request()
        except TypeError as e:
            if e.__str__().startswith("'NoneType'"):
                pass
            else:
                raise TypeError(e)
        if not token:
            try:
                token = self._kernel.user.token
            except Exception as e:
                raise(e)
        return token


from __future__ import annotations

import firefly as ff

from firefly_iaaa import application, domain


class AuthorizeRequest(ff.DomainService):
    _authorize_request: domain.OAuthAuthorizeRequest = None
    _message_factory: ff.MessageFactory = None
    _add_method_to_headers: domain.AddMethodToHeaders = None

    def __call__(self, **kwargs):
        if 'access_token' not in kwargs:
            return False
        return self._authorize_request(self._make_message(kwargs), **kwargs)

    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs)
        message_body = {
            'headers': headers,
            'access_token': incoming_kwargs.get('access_token'),
            
        }

        for field in ('username', 'password', 'scopes'):
            if incoming_kwargs.get(field):
                message_body[field] = incoming_kwargs.get(field)

        return self._message_factory.query(
            name='OauthAuthorizationRequestMessage',
            data=message_body
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint


class GenericOauthIamEndpoint(GenericOauthEndpoint):

    def __call__(self, **kwargs):
        pass

    def _make_local_response(self, headers, body):
        tokens = body.get('tokens')
        cookies = []
        access_cookie = {
            'name': 'accessToken',
            'value': tokens['access_token'],
            'httponly': True,
            'max_age': tokens['expires_in'],
        }
        cookies.append(access_cookie)
        if 'refresh_token' in tokens:
            refresh_cookie = {
                'name': 'refreshToken',
                'value': tokens['refresh_token'],
                'httponly': True,
            }
            cookies.append(refresh_cookie)
        envelope = self._make_response(tokens, headers=headers, cookies=cookies)
        return envelope



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from firefly_iaaa.domain.service.add_method_to_headers import AddMethodToHeaders
from firefly_iaaa.domain.service.authorize_request import AuthorizeRequest
from firefly_iaaa.domain.service.create_token import CreateToken
from firefly_iaaa.domain.service.cognito_login import CognitoLogin
from firefly_iaaa.domain.service.decode_token import DecodeToken
from firefly_iaaa.domain.service.get_client_id import GetClientId
from firefly_iaaa.domain.service.make_client_user_entities import MakeClientUserEntities
from firefly_iaaa.domain.service.middleware import *
from firefly_iaaa.domain.service.oauth_login import OAuthLogin
from firefly_iaaa.domain.service.oauth_provider import OauthProvider
from firefly_iaaa.domain.service.oauth_register import OAuthRegister
from firefly_iaaa.domain.service.request_validator import OauthRequestValidators
from firefly_iaaa.domain.service.send_reset_email import SendResetEmail



from __future__ import annotations

import firefly as ff
from jwt import ExpiredSignatureError

from firefly_iaaa import domain


@ff.query_handler()
class DecodedToken(ff.ApplicationService):
    _decode_token: domain.DecodeToken = None

    def __call__(self, token: str, audience: str = None, **kwargs):
        try:
            return self._decode_token(token, audience)
        except ExpiredSignatureError:
            return {'message': 'error', 'error': 'Expired Token'}



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
from __future__ import annotations
from datetime import datetime, timedelta
from operator import itemgetter
import uuid

import firefly as ff
from firefly_iaaa import domain
import jwt
from oauthlib.oauth2 import Server
from oauthlib.common import Request

from .request_validator import OauthRequestValidators


class OauthProvider(ff.DomainService):
    _cache: ff.Cache = None
    _secret_key: str = None
    _issuer: str = None
    _registry: ff.Registry = None

    def __init__(self, validator: OauthRequestValidators):
        self._server = Server(
            validator,
            token_generator=(lambda x: self.generate_token(x, 'access_token')),
            refresh_token_generator=(lambda x: self.generate_token(x, 'refresh_token')),
            token_expires_in=3600
        )

    def generate_token(self, request, token_type):
        scopes = request['scopes'] if isinstance(request, dict) else request.scopes

        token = {
            'jti': str(uuid.uuid4()),
            'aud': request.client_id,
            'iss': self._issuer,
            'scope': ' '.join(self._convert_from_scopes_to_string(scopes)),
            'tenant': request.user.tenant_id
        }
        if request.user:
            token['sub'] = request.user.sub
        if token_type == 'access_token':
            token['exp'] = datetime.utcnow() + timedelta(seconds=request.expires_in)
        token = jwt.encode(token, self._secret_key, algorithm='HS256')
        return token

    def validate_pre_auth_request(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        scopes, credentials = self._server.validate_authorization_request(uri, http_method, body, headers)

        credentials_key = str(uuid.uuid4())
        credentials = self._break_down_credentials(credentials)

        self._cache.set(credentials_key, value=credentials, ttl=180)

        return self._convert_from_scopes_to_string(scopes), credentials, credentials_key

    def validate_post_auth_request(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        credentials_key = body.get('credentials_key')
        if not credentials_key:
            raise ff.UnauthorizedError('Missing credentials key')

        credentials = self._cache.get(credentials_key)
        if not credentials:
            raise ff.UnauthorizedError('Credentials expired or invalid')

        credentials = self._build_up_credentials(credentials)

        if not request.scopes:
            raise ff.UnauthorizedError('Missing scopes')
        headers, body, status = self._server.create_authorization_response(uri, http_method, body, headers, scopes=request.scopes, credentials=credentials)

        if headers.get('Location'):
            self._cache.delete(credentials_key)
        return headers, body, status

    def create_token_response(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)

        headers, body, status = self._server.create_token_response(uri, http_method, body, headers)
        return headers, body, status

    def create_response(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        return self._server.create_authorization_response(
            uri, http_method, body, headers
        )

    def verify_request(self, request: ff.Message, scopes):
        uri, http_method, body, headers = self._get_request_params(request)

        is_valid, req = self._server.verify_request(uri, http_method, body, headers, scopes=scopes)
        req = self.scrub_sensitive_data(req)
        return is_valid, req

    def create_introspect_response(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        headers, body, status = self._server.create_introspect_response(uri, http_method, body, headers)
        return headers, body, status

    def create_revocation_response(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        headers, body, status = self._server.create_revocation_response(uri, http_method, body, headers)
        return headers, body, status

    def authenticate_client(self, request: ff.Message):
        uri, http_method, body, headers = self._get_request_params(request)
        oauth_request = Request(uri, http_method, body, headers)
        return self._server.request_validator.authenticate_client(oauth_request) #!! Check

    @staticmethod
    def _get_request_params(request: ff.Message):
        uri = request.headers.get('Origin') or request.headers.get('origin') or request.headers.get('Referer') or request.headers.get('uri') or ''
        http_method = request.headers.get('method') or request.headers.get('http_method')
        body = request.to_dict()
        headers = request.headers
        return uri, http_method, body, headers

    def _break_down_credentials(self, credentials):
        credentials['request'] = self.scrub_sensitive_data(credentials['request']).__dict__
        return credentials

    def _build_up_credentials(self, credentials):
        values = ('uri', 'http_method', 'body', 'headers')
        uri, http_method, body, headers = itemgetter('uri', 'http_method', 'body', 'headers')(credentials['request'])

        request = Request(uri, http_method, body, headers)

        for k, v in credentials['request'].items():
            if k not in values:
                setattr(request, k, v)
        credentials['request'] = self._add_entities_to_credentials(request)

        return credentials

    @staticmethod
    def scrub_sensitive_data(request: Request):
        try:
            request.client = request.client.generate_scrubbed_client()
        except AttributeError:
            try:
                request['client'] = request['client'].generate_scrubbed_client()
            except (KeyError, TypeError):
                pass
        try:
            request.user = request.user.generate_scrubbed_user()
        except AttributeError:
            try:
                request['user'] = request['user'].generate_scrubbed_user()
            except (KeyError, TypeError):
                pass
        try:
            request.tenant = request.tenant.id
        except AttributeError:
            try:
                request['tenant'] = request['tenant'].id
            except (KeyError, TypeError):
                pass
        return request

    def _add_entities_to_credentials(self, request: Request):
        try:
            client = self._registry(domain.Client).find(lambda x: x.client_id == request.client['client_id'])
            request.client = client
        except AttributeError:
            try:
                client = self._registry(domain.Client).find(lambda x: x.client_id == request['client']['client_id'])
                request['client'] = client
            except (KeyError, TypeError):
                pass
        try:
            user = self._registry(domain.User).find(lambda x: x.sub == request.user['sub'])
            request.user = user
        except (AttributeError, TypeError):
            try:
                user = self._registry(domain.User).find(lambda x: x.sub == request['user']['sub'])
                request['user'] = user
            except (KeyError, TypeError):
                pass
        try:
            tenant = self._registry(domain.Tenant).find(lambda x: x.id == request.tenant)
            request.tenant = tenant
        except (AttributeError, TypeError):
            try:
                tenant = self._registry(domain.Tenant).find(lambda x: x.id == request['tenant'])
                request['tenant'] = tenant
            except (KeyError, TypeError):
                pass
        return request

    @staticmethod
    def _convert_from_scopes_to_string(scopes):
        converted = []
        for scope in scopes:
            if isinstance(scope, str):
                #  Checking if the scope might be 
                scope = scope.split(' ')[-1]
                if scope.endswith(']'):
                    converted.append(scope[1:-2])
                else:
                    converted.append(scope)
            elif isinstance(scope, dict):
                converted.append(scope['id'])
            else:
                converted.append(scope.id)
            
        return converted



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
from __future__ import annotations
from typing import Any
import firefly as ff
from datetime import datetime, timedelta


class MockCache(ff.Cache):
    _storage: dict = {}

    def set(self, key: str, value: Any, ttl: int = None, **kwargs):
        time = (datetime.now() + timedelta(seconds=ttl)) if ttl else None
        self._storage[key] = {'value': value, 'ttl': time}
    
    def get(self, key: str, **kwargs):
        item = self._storage.get(key)
        if not item:
            return None
        if item['ttl'] is None or datetime.now() < item['ttl']:
            return item['value']
        return None

    def delete(self, key: str, **kwargs):
        self._storage[key] = None
        del self._storage[key]
        return None

    def clear(self, **kwargs):
        return None

    def increment(self, key: str, amount: int = 1, **kwargs) -> Any:
        return None

    def decrement(self, key: str, amount: int = 1, **kwargs) -> Any:
        return None

    def add(self, key: str, value: Any, **kwargs) -> Any:
        return None

    def remove(self, key: str, value: Any, **kwargs) -> Any:
        return None

    def list(self):
        return self._storage.items()


from __future__ import annotations

import firefly as ff
import json

from firefly_iaaa import domain
from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint


class CreateToken(ff.DomainService):
    _oauth_provider: domain.OauthProvider = None
    _message_factory: ff.MessageFactory = None
    _get_client_id: domain.GetClientId = None
    _add_method_to_headers: domain.AddMethodToHeaders = None

    def __call__(self, passed_in_kwargs:dict, **kwargs):
        message = self._make_message(passed_in_kwargs)

        headers, body, status = self._oauth_provider.create_token_response(message)

        return [headers, json.loads(body)]

    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs)
        message_body = {
            'headers': headers,
            'grant_type': incoming_kwargs.get('grant_type'),
            'client_id': self._get_client_id(incoming_kwargs.get('client_id')),
            'state': incoming_kwargs.get('state')
        }

        for field in ('username', 'password', 'client_secret', 'code', 'code_verifier', 'refresh_token', 'scopes', 'scope'):
            if incoming_kwargs.get(field):
                message_body[field] = incoming_kwargs.get(field)

        return self._message_factory.query(
            name='OauthCreateTokenMessage',
            data=message_body
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
from typing import Dict
import os

import firefly as ff
import firefly_iaaa.domain as domain


class OAuthLogin(ff.DomainService, ff.LoggerAware):
    _cognito_login: domain.CognitoLogin = None
    _registry: ff.Registry = None
    _oauth_register: domain.OAuthRegister = None
    _create_token: domain.CreateToken = None

    def __call__(self, passed_in_kwargs: dict):
        self.debug('Logging in with In-House')
        username = passed_in_kwargs['username']
        password = passed_in_kwargs['password']
        tokens = [None, None]

        found_user: domain.User = self._registry(domain.User).find(lambda x: x.email == username)
        if found_user:
            passed_in_kwargs['grant_type'] = 'password'
            self.info('We found a user, trying to login with password')
            if found_user.correct_password(password):
                passed_in_kwargs = self._set_client_id(found_user, passed_in_kwargs)
                tokens = self._get_tokens(passed_in_kwargs)
                resp = [tokens[0], {'tokens': tokens[1], 'user': found_user.generate_scrubbed_user()}]
            else:
                resp = self._try_cognito(username, password)
                self.info('User password incorrect, trying Cognito in case password null')
                if resp is None:
                    raise ff.UnauthenticatedError('Incorrect username/password combination')
        else:
            self.info('No user exists, trying Cognito')
            resp = self._try_cognito(username, password)

        if resp is None:
            raise ff.NotFound()

        return resp


    def _try_cognito(self, username: str, password: str):
        self.debug('Switching to Cognito Log in')
        if self._registry(domain.User).find(lambda x: x.email == username) is None:
            try:
                resp = self._cognito_login(username, password) #data has tokens and idToken
                message, error, success, data = resp.values()
                if error:
                    if message:
                        ff.UnauthenticatedError(message)
                    else:
                        ff.UnauthenticatedError(error)
                if success:
                    resp = self._transfer_cognito_user_to_native_user(username, password, data['decoded_id_token'])
                    return resp
            except Exception as e:
                raise ff.UnauthenticatedError()
        else:
            raise ff.UnauthenticatedError('Incorrect Password')
        

    def _transfer_cognito_user_to_native_user(self, username: str, password: str, data: Dict):
        self.debug('Transfering Cognito user to In-House user')
        data['email'] = username
        data['username'] = username
        data['password'] = password
        resp = self._oauth_register(data)
        if resp[1]['tokens']:
            return resp
        raise Exception('Something went wrong')

    def _get_tokens(self, kwargs: dict):
        # kwargs = self._set_referer(kwargs)
        resp = self._create_token(kwargs)
        return resp

    def _set_client_id(self, found_user, kwargs):
        if not kwargs.get('client_id'):
            user_client = self._registry(domain.Client).find(lambda x: x.tenant_id == found_user.tenant_id)
            kwargs['client_id'] = user_client.client_id
        return kwargs

    def _set_referer(self, kwargs: dict):
        headers = {
            'http_request': {
                'headers': {
                    'Referer': 'https://www.pwrlab.com/'
                }
            }
        }
        if not kwargs.get('headers'):
            kwargs['headers'] = headers
        elif not kwargs['headers'].get('http_request'):
            kwargs['headers']['http_request'] = headers['http_request']
        elif not kwargs['headers']['http_request'].get('headers'):
            kwargs['headers']['http_request']['headers'] = headers['http_request']['headers']
        elif not kwargs['headers']['http_request']['headers'].get('Referer'):
            kwargs['headers']['http_request']['headers']['Referer'] = headers['http_request']['headers']['Referer']
        return kwargs


from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain

import sys
import importlib.util


if 'firefly_aws' in sys.modules or importlib.util.find_spec('firefly_aws') is not None:

    class BaseAwsTokenGenerationService:
        _registry: ff.Registry = None

        def _get_token_access_rights(self, event: dict):
            user: domain.User = self._registry(domain.User).find(event['request']['userAttributes']['sub'])
            if user is None:
                self.info('No record for user "%s"', event['request']['userAttributes']['sub'])
                return event

            scopes = []
            for role in user.roles:
                scopes.extend(list(map(str, role.scopes)))

            event['response'] = {
                'claimsOverrideDetails': {
                    'groupOverrideDetails': {
                        'groupsToOverride': scopes,
                    }
                }
            }

            return event

    @ff.command_handler('firefly_aws.TokenGeneration_HostedAuth')
    class HandleHostedAuth(ff.ApplicationService, BaseAwsTokenGenerationService):
        _registry: ff.Registry = None

        def __call__(self, event: dict, **kwargs):
            return self._get_token_access_rights(event)


    @ff.command_handler('firefly_aws.TokenGeneration_RefreshTokens')
    class HandleRefreshTokens(ff.ApplicationService, BaseAwsTokenGenerationService):
        _registry: ff.Registry = None

        def __call__(self, event: dict, **kwargs):
            return self._get_token_access_rights(event)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
from firefly_iaaa.application.api.generic_oauth_iam_endpoint import GenericOauthIamEndpoint
import firefly_iaaa.domain as domain


@ff.rest('/iaaa/login', method='POST', tags=['public'], secured=False)
class OAuthLogin(GenericOauthIamEndpoint):
    _oauth_login: domain.OAuthLogin = None
    _kernel: ff.Kernel = None

    def __call__(self, **kwargs):
        self.debug('Logging in with In-House')
        kwargs = self._fix_email(kwargs)
        if 'username' not in kwargs or 'password' not in kwargs:
            return self._make_error_response('Missing username/password')

        resp = self._oauth_login(kwargs)
        
        if 'error' in resp:
            return self._make_error_response(resp)

        return self._make_local_response(*resp)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List
from datetime import datetime

import firefly as ff

class Cohort(ff.AggregateRoot):
    id: str = ff.id_()
    tenant: str = ff.required(index=True)
    name: str = ff.required(length=256, index=True)
    status: str = ff.required(str)
    available_permissions: List[str] = ff.list_()
    active: bool = ff.required(default=True)
    start_date: datetime = ff.optional(default=None)
    end_date: datetime = ff.optional(default=None)




from __future__ import annotations

import firefly as ff
import urllib
import ast
import base64
import firefly_iaaa.domain as domain
from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint

#! Secured?
@ff.rest('/iaaa/authorize', method='GET', tags=['public'], secured=False)
class OauthAuthorizationRequestService(GenericOauthEndpoint):
    _registry: ff.Registry = None
    _subdomain: str = None

    def __call__(self, **kwargs):
        kwargs = self._fix_email(kwargs)
        message = self._make_message(kwargs) #! check more

        resp = self._oauth_provider.validate_pre_auth_request(message)
        return self._make_response(forwarding_address=self._make_local_response(*resp))

    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs, 'GET')
        message_body = {
            'headers': headers,
            'client_id': self._get_client_id(incoming_kwargs.get('client_id')),
            'state': incoming_kwargs.get('state'),
            'response_type': incoming_kwargs.get('response_type'),
            'code_challenge': incoming_kwargs.get('code_challenge'),
        }

        if incoming_kwargs.get('username'):
            message_body['username'] = incoming_kwargs.get('username') 
        if incoming_kwargs.get('password'):
            message_body['password'] = incoming_kwargs.get('password') 
        if incoming_kwargs.get('client_secret'):
            message_body['client_secret'] = incoming_kwargs.get('client_secret')
        if incoming_kwargs.get('cohort_id'):
            message_body['cohort_id'] = incoming_kwargs.get('cohort_id')

        return self._message_factory.query(
            name='OauthAuthorizationRequestMessage',
            data=message_body
        )

    def _get_cohort_name(self, cohort_id: str):
        if cohort_id:
            cohort = self._registry(domain.Cohort).find(lambda x: x.id == cohort_id)
            if cohort:
                return cohort.name
        return None


    def _make_local_response(self, scopes, credentials, credentials_key):
        cohort_name = self._get_cohort_name(credentials['request']['body'].get('cohort_id'))
        resp = {
            'redirect_uri': credentials.get('redirect_uri'),
            'client_id': credentials.get('client_id'),
            'scopes': scopes,
            'credentials_key': credentials_key,
            'response_type': credentials.get('response_type'),
            'client': {'tenant': credentials['request']['client']['tenant_name'], 'cohort': cohort_name or ''}
        }
        if 'code_challenge' in credentials:
            resp['code_challenge'] = credentials['code_challenge']
        if 'code_challenge_method' in credentials:
            resp['code_challenge_method'] = credentials['code_challenge_method']
        # if 'claims' in credentials:
        #     kwargs['claims'] = json.dumps(credentials['claims'])
        redirect_url = f'https://{self._subdomain}.pwrlab.com/authorize?' #!!!!

        for k, v in resp.items():
            if not v:
                continue
            if k == 'redirect_uri':
                v = base64.b64encode(v.encode('utf-8'))
            val = urllib.parse.quote(str(v), safe='')
            redirect_url += f'{k}={val}&'
        return redirect_url[0:-1]


#! Secured?
@ff.rest(
    '/iaaa/authorize', method='POST', tags=['public'], secured=False
)
class OauthCreateAuthorizationService(GenericOauthEndpoint):

    def __call__(self, **kwargs):
        kwargs = self._fix_email(kwargs)
        message = self._make_message(kwargs) #! check more
        
        access_token = self._grab_token_from_headers()

        self._set_user_from_token(access_token, message.client_id)

        headers, body, status = self._oauth_provider.validate_post_auth_request(message)
        if not headers and not body and not status:
            raise ff.UnauthorizedError()
        if status == 302 and 'Location' in headers:
            body = body or {}
            body['redirect'] = headers['Location']
        return self._make_response(body)


    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs)
        redirect_uri = self._get_redirect_uri(incoming_kwargs)
        message_body = {
            'headers': headers,
            'state': incoming_kwargs.get('state'),
            'redirect_uri': redirect_uri,
            'client_id': incoming_kwargs.get('client_id'),
            'scopes': incoming_kwargs.get('scopes'),
            'credentials_key': incoming_kwargs.get('credentials_key'),
            'response_type': incoming_kwargs.get('response_type')
        }


        return self._message_factory.query(
            name='OauthCreateAuthorizationMessage',
            data=message_body
        )

    @staticmethod
    def _get_redirect_uri(incoming_kwargs: dict):

        redirect_uri = incoming_kwargs.get('redirect_uri')
        if not redirect_uri:
            raise ff.UnauthorizedError('Missing redirect_uri')
        redirect_uri = urllib.parse.unquote(redirect_uri)
        redirect_uri = base64.b64decode(ast.literal_eval(redirect_uri))
        return redirect_uri.decode('utf-8')

    def _grab_token_from_headers(self):
        for k, v in self._kernel.http_request['headers'].items():
            if k.lower() == 'authorization':
                if not v.lower().startswith('bearer'):
                    raise ff.UnauthorizedError()
                return v.split(' ')[-1]
        raise ff.UnauthorizedError()

    def _set_user_from_token(self, access_token: str, client_id: str):
        bearer_token = self._registry(domain.BearerToken).find(lambda x: x.access_token == access_token)
        client = self._registry(domain.Client).find(lambda x: x.client_id == client_id)
        if not client or not bearer_token:
            ff.UnauthorizedError()
        if not bearer_token.validate_access_token(access_token, client):
            ff.UnauthorizedError('Invalid access token')
        user = bearer_token.user
        decoded_token = self._decode_token(access_token)
        if not decoded_token:
            ff.UnauthorizedError('Invalid access_token/client combination')
        self._kernel.user = ff.User(id=user.sub, scopes=decoded_token['scope'].split(' '), tenant=user.tenant_id, token=decoded_token)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
from .generic_oauth_middleware import GenericOauthDomainMiddleware


class OAuthAuthenticator(GenericOauthDomainMiddleware):

    def __call__(self, message: ff.Message, **kwargs):
        self.info('Authenticating')
        message = self._fix_email(message)
        self.info(self._kernel)
        if self._kernel.http_request and self._kernel.secured:
            token = self._retrieve_token_from_http_request()
            if token:
                token = token.split(' ')[-1]
            if token is None:
                try:
                    token = message.access_token
                except:
                    raise ff.UnauthenticatedError()

            self.debug('Decoding token')
            try:
                decoded = self._decode_token(token, self._kernel.user.id)
            except:
                raise ff.UnauthenticatedError()
    
            self._kernel.user.token = decoded
            self._kernel.user.scopes = decoded['scope'].split(' ')
            return True
        return self._kernel.secured is not True



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import os

import firefly as ff
import uuid
import firefly_iaaa.domain as domain


class MakeClientUserEntities(ff.DomainService):
    _registry: ff.Registry = None
    _context_map: ff.ContextMap = None
    _context: str = None

    def __call__(self, username: str, password: str, tenant_name: str, **kwargs):
        module_name = self._context_map.\
            get_context(self._context).\
            config.\
            get('domain_module', '{}.domain')
        module = importlib.import_module(module_name.format(self._context))
        user_entity = module.__dict__.get('User')
        role_entity = module.__dict__.get('Role')
        tenant_entity = module.__dict__.get('Tenant')
        scope_entity = module.__dict__.get('Scope')
        client_entity = module.__dict__.get('Client', domain.Client)
        
        roles = []
        if 'roles' in kwargs:
            for role in kwargs['roles']:
                r = self._registry(role_entity).find(lambda x: x.name == role)
                roles.append(r)
        else:
            r = self._registry(role_entity).find(lambda x: x.name == 'Consumer App User')
            roles.append(r)
        kwargs['roles'] = roles

        kwargs['email'] = kwargs.get('email', username)

        if kwargs['grant_type'] == 'password':
            kwargs['client_id'], kwargs['tenant'] = self._get_consumer_client(client_entity)
            user = user_entity.create(
                username=username,
                password=password,
                **kwargs
            )
            self._registry(user_entity).append(user)
            return user
        tenant = tenant_entity(
            name=kwargs['tenant_name']
        )
        kwargs['tenant'] = tenant
        user = user_entity.create(
            username=username,
            password=password,
            **kwargs
        )

        kwargs['client_id'] = user.sub
        kwargs['name'] = kwargs.get('name', tenant_name)
        kwargs = self._make_params(kwargs)

        if kwargs.get('scopes') is None or len(kwargs['scopes']) == 0:
            kwargs['scopes'] = ['iaaa.default.read']
        scopes = []
        for scope in kwargs['scopes']:
            s = self._registry(scope_entity).find(scope)
            scopes.append(s)
        kwargs['scopes'] = scopes

        client = client_entity.create(**kwargs)

        # Append at end to avoid appending before an error during entity creation
        self._registry(tenant_entity).append(tenant)
        self._registry(user_entity).append(user)
        self._registry(client_entity).append(client)
        return user

    def _make_params(self, kwargs: dict):

        self._validate_base_params(kwargs)
        grant_type = kwargs['grant_type']
        if grant_type == 'authorization_code':
            kwargs = self._add_auth_code_params(kwargs, False)
        elif grant_type == 'authorization_code_w_pkce':
            kwargs = self._add_auth_code_params(kwargs)
            kwargs['grant_type'] = 'authorization_code'
        elif grant_type == 'implicit':
            kwargs = self._add_auth_code_params(kwargs)
            kwargs['allowed_response_types'] = 'token'
        elif grant_type == 'client_credentials':
            kwargs['client_secret'] = str(uuid.uuid4())
        elif grant_type == 'password':
            name = kwargs['name']
            if name.startswith('user_tenant'):
                name = name[11:]
            kwargs['name'] = f'user_client_{name}'

        else:
            raise Exception('Invalid grant type')

        return kwargs

    def _validate_base_params(self, kwargs: dict):
        fields = ['scopes', 'grant_type']
        self._check_kwargs_for_fields(fields, kwargs)

    def _add_auth_code_params(self, kwargs: dict, uses_pkce: bool = True):
        fields = ['default_redirect_uri', 'redirect_uris']
        self._check_kwargs_for_fields(fields, kwargs)
        kwargs.update({
            'allowed_response_types': 'code',
            'uses_pkce': uses_pkce,
        })
        return kwargs

    def _get_consumer_client(self, client_entity):
        client: client_entity = self._registry(client_entity).find(lambda x: x.client_id == os.environ['CONSUMER_CLIENT_ID'])
        tenant = client.tenant
        return client.client_id, tenant

    def _check_kwargs_for_fields(self, fields, kwargs):
        for field in fields:
            if field not in kwargs:
                raise Exception(f'Missing required field: {field}')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain
from firefly_iaaa.application.api.generic_endpoint import GenericEndpoint


@ff.rest('/iaaa/change-password', method='POST', tags=['public'], secured=False)
class ChangePassword(GenericEndpoint):
    _cache: ff.Cache = None

    def __call__(self, **kwargs):
        self.debug('Changing password for User')
        try:
            request_id = kwargs['request_id']
            payload = self._cache.get(request_id)
            if not payload:
                return self._make_error_response('Password request is invalid')
            if payload['message'] == 'reset':
                username = payload['username'].lower()
            new_password = kwargs['new_password']
        except KeyError as e:
            return self._make_error_response('Missing password')

        found_user: domain.User = self._registry(domain.User).find(lambda x: x.email == username)

        if found_user:
            found_user.change_password(new_password)
            self.debug('Password Successfully Changed')
            self._cache.delete(request_id)
            return self._make_response()
        return self._make_error_response('No User found')



from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain


@ff.command_handler()
class AddRoleToUser(ff.ApplicationService):
    _registry: ff.Registry = None

    def __call__(self, sub: str, role_id: str = None, role_name: str = None, **kwargs):
        if role_id is None and role_name is None:
            raise ff.ApiError('must provide one of role_id or role_name')

        if role_id is not None:
            role = self._registry(domain.Role).find(role_id)
        else:
            role = self._registry(domain.Role).find(lambda r: r.name == role_name)

        if not role:
            raise ff.NotFound('role not found')

        user = self._registry(domain.User).find(sub)
        if not user:
            # TODO If firefly-aws is installed, attempt to pull the user from cognito.
            user = domain.User(sub=sub)
            self._registry(domain.User).append(user)

        user.roles.append(role)



from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain
from firefly_iaaa.application.api.generic_endpoint import GenericEndpoint


class GenericOauthEndpoint(GenericEndpoint):
    _oauth_provider: domain.OauthProvider = None
    _kernel: ff.Kernel = None
    _registry: ff.Registry = None
    _message_factory: ff.MessageFactory = None
    _get_client_id: domain.GetClientId = None
    _add_method_to_headers: domain.AddMethodToHeaders = None

    def __call__(self, **kwargs):
        pass

    def _make_message(self, incoming_kwargs: dict):
        pass

    def _fix_email(self, kwargs):
        if kwargs.get('username'):
            kwargs['username'] = kwargs['username'].lower()
        email = kwargs.get('email', kwargs.get('username'))
        if email:
            kwargs['email'] = email.lower()
        return kwargs


#  Copyright (c) 2020 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

import firefly as ff
from firefly import domain as ffd
from firefly_iaaa.domain.service.decode_token import DecodeToken
import boto3


class CognitoLogin(ff.DomainService, ff.LoggerAware):
    _decode_token: DecodeToken = None
    _kernel: ffd.Kernel = None


    def __call__(self, username: str, password: str):
        client = boto3.client('cognito-idp')

        resp, msg = self._initiate_auth(client, username, password)

        if msg != None:
            return {'message': msg, 
                    'error': 'No user exists', 'success': False, 'data': None}
        if resp.get('AuthenticationResult'):
            return {
                'message': 'success', 
                'error': '',
                'success': True, 
                'data': {
                    'id_token': resp['AuthenticationResult']['IdToken'],
                    'refresh_token': resp['AuthenticationResult']['RefreshToken'],
                    'access_token': resp['AuthenticationResult']['AccessToken'],
                    'expires_in': resp['AuthenticationResult']['ExpiresIn'],
                    'token_type': resp['AuthenticationResult']['TokenType'],
                    'decoded_id_token': self._decode_token(resp['AuthenticationResult']['IdToken'])
                }
            }
        else:
            return {
                'error': 'Something went wrong, no authentication results', 
                'success': False, 
                'data': None,
                'message': 'error'
            }


    def _initiate_auth(self, client, username, password):
        try:
            resp = client.initiate_auth(
                        ClientId=os.environ['CLIENT_ID'],
                        AuthFlow='USER_PASSWORD_AUTH',
                        AuthParameters={
                            'USERNAME': username,
                            'PASSWORD': password,
                        },
                        ClientMetadata={
                        'username': username,
                        'password': password,
                    })
        except client.exceptions.NotAuthorizedException:
            return None, 'The username or password is incorrect'
        except client.exceptions.UserNotConfirmedException:
            return None, 'User is not confirmed'
        except KeyError as e:
            return None, f'Key Error: {e.__str__()}'
        except Exception as e:
            return None, e.__str__()
        return resp, None



from __future__ import annotations
import json

import firefly as ff
import firefly_iaaa.domain as domain
from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint


@ff.rest(
    '/iaaa/revoke', method='POST', tags=['public'], secured=True
)
class OauthTokenRevocationService(GenericOauthEndpoint):
    _oauth_provider: domain.OauthProvider = None
    _kernel: ff.Kernel = None
    _message_factory: ff.MessageFactory = None

    def __call__(self, **kwargs):
        kwargs = self._fix_email(kwargs)
        message = self._make_message(kwargs)

        headers, body, status =  self._oauth_provider.create_revocation_response(message)

        return self._make_response(headers=headers)

    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs)
        message_body = {
            'headers': headers,
            'token': incoming_kwargs.get('token'),
            "client_id": self._get_client_id(incoming_kwargs.get('client_id')),
            "state": incoming_kwargs.get('state')
        }

        if incoming_kwargs.get('username'):
            message_body['username'] = incoming_kwargs.get('username') 
        if incoming_kwargs.get('password'):
            message_body['password'] = incoming_kwargs.get('password') 
        if incoming_kwargs.get('client_secret'):
            message_body['client_secret'] = incoming_kwargs.get('client_secret')


        return self._message_factory.query(
            name='OauthRevokeTokenMessage',
            data=message_body
        )



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List
import uuid

import firefly as ff
from .tenant import Tenant
from .role import Role
from .scope import Scope

authorization_code = 'authorization_code'
implicit = 'implicit'
resource_owner_password_credentials = 'password'
client_credentials = 'client_credentials'
refresh = 'refresh_token'


class Client(ff.AggregateRoot):
    id: str = ff.id_() 
    client_id: str = ff.optional(validators=[ff.HasLength(36)], index=True) # Needs to have 'client_id', should be same as sub if user/cleint combo
    external_id: str = ff.optional(index=True)
    name: str = ff.required()
    grant_type: str = ff.required(validators=[ff.IsOneOf((
        authorization_code, implicit, resource_owner_password_credentials, client_credentials, refresh
    ))])
    default_redirect_uri: str = ff.optional()
    redirect_uris: List[str] = ff.list_()
    scopes: List[Scope] = ff.list_()
    roles: List[Role] = ff.list_()
    allowed_response_types: List[str] = ff.list_(validators=[ff.IsOneOf(('code', 'token'))])
    uses_pkce: bool = ff.optional(default=True)
    client_secret: str = ff.optional(str, length=36)
    is_active: bool = True
    tenant: Tenant = ff.optional(index=True)
    tenant_id: str = ff.optional(index=True)

    @classmethod
    def create(cls, **kwargs):
        try:
            kwargs['tenant_id'] = kwargs['tenant'].id
        except KeyError:
            raise ff.MissingArgument('Tenant is a required field for Client::create()')
        try:
            kwargs['grant_type'] = kwargs['grant_type']
        except KeyError:
            raise ff.MissingArgument('Grant Type is a required field for Client::create()')
        if not kwargs.get('client_id'):
            kwargs['client_id'] = str(uuid.uuid4())
        kwargs['id'] = kwargs['client_id']
        return cls(**ff.build_argument_list(kwargs, cls))

    def validate_redirect_uri(self, redirect_uri: str):
        return redirect_uri in self.redirect_uris

    def validate_response_type(self, response_type: str):
        return response_type in self.allowed_response_types

    def validate_grant_type(self, grant_type: str):
        return grant_type in (self.grant_type, refresh)

    def validate_scopes(self, scopes: List[str]):
        client_scopes = self.get_scopes()
        if isinstance(scopes, str):
            scopes = scopes[1:-1].replace("'", '').replace(',', '').split(' ')
        #  Include ANY scopes
        # if not scopes:
        #     return False
        # for scope in scopes:
        #     if scope in client_scopes:
        #         return True
        # return False

        #  Include ALL scopes
        if not scopes:
            return False
        for scope in scopes:
            if scope not in client_scopes:
                return False
        return True

    def validate(self):
        return self.is_active

    def requires_pkce(self):
        return self.uses_pkce

    def is_confidential(self):
        return self.grant_type in (client_credentials, resource_owner_password_credentials) or \
            (self.grant_type == authorization_code and self.requires_pkce())

    def validate_client_secret(self, secret):
        return self.client_secret == secret

    def inactivate(self):
        self.is_active = False

    def add_role(self, role: Role):
        if isinstance(role, Role):
            self.roles.append(role)

    def generate_scrubbed_client(self):
        return {
            'client_id': self.client_id,
            'external_id': self.external_id,
            'name': self.name,
            'grant_type': self.grant_type,
            'default_redirect_uri': self.default_redirect_uri,
            'redirect_uris': self.redirect_uris,
            'scopes': self.get_scopes(),
            'allowed_response_types': self.allowed_response_types,
            'is_active': self.is_active,
            'tenant_id': self.tenant_id,
            'tenant_name': self.tenant.name,
        }

    def _get_entity_scopes(self):
        roles = [scope for role in self.roles for scope in role.scopes]
        roles += self.scopes
        return roles

    def get_scopes(self):
        return [scope.id for scope in self._get_entity_scopes()]

    def _get_scopes_from_roles(self):
        roles = []
        for role in self.roles:
            if isinstance(role, Role):
                roles.append(*role.scopes)
            else:
                roles.append(*role['scopes'])
        return roles



from __future__ import annotations

import firefly as ff

from firefly_iaaa import domain

@ff.query_handler()
class AuthorizeRequest(ff.ApplicationService):
    _authorize_request: domain.AuthorizeRequest = None

    def __call__(self, token: str, **kwargs):
        kwargs['access_token'] = token
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['uri'] = ''
        return self._authorize_request(**kwargs)



from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain


class GetClientId(ff.DomainService):
    _kernel: ff.Kernel = None
    _registry: ff.Registry = None

    def __call__(self, client_id: str, **kwargs):
        if client_id:
            return client_id
        user = self._registry(domain.User).find(lambda x: x.sub == self._kernel.user.id)
        if not user:
            return self._kernel.user.id
        client = self._registry(domain.Client).find(lambda x: x.tenant_id == user.tenant_id)
        if client is None:
            client = self._registry(domain.Client).find(lambda x: x.client_id == user.sub)
        return client.client_id


#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from datetime import datetime
from typing import List

import firefly as ff
from firefly_iaaa.domain.entity.client import Client
from firefly_iaaa.domain.entity.user import User
from firefly_iaaa.domain.entity.scope import Scope


class AuthorizationCode(ff.AggregateRoot):
    id_: str = ff.id_()
    client: Client = ff.required(index=True)
    user: User = ff.required()
    scopes: List[Scope] = ff.list_()
    redirect_uri: str = ff.optional()
    claims: dict = ff.optional()
    code: str = ff.required(str, length=36, index=True)
    expires_at: datetime = ff.required()
    state: str = ff.required()
    challenge: str = ff.optional(str, length=128)
    challenge_method: str = ff.optional(validators=[ff.IsOneOf(('S256', 'plain'))])
    claims: dict = ff.optional()
    verifier: str = ff.optional()
    is_valid: bool = True

    def validate_redirect_uri(self, redirect_uri: str):
        return self.redirect_uri == redirect_uri

    def invalidate(self):
        self.is_valid = False

    def is_expired(self):
        return self.expires_at < datetime.utcnow()

    def validate(self, client_id: str):
        return self.is_valid and client_id == self.client.client_id and not self.is_expired()

    def get_scopes(self):
        return [scope.id for scope in self.scopes]



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from datetime import datetime, date
from typing import List

import bcrypt
import firefly as ff

from firefly_iaaa.domain.entity.role import Role
from firefly_iaaa.domain.entity.tenant import Tenant
from firefly_iaaa.domain.value_object.address import Address


class User(ff.AggregateRoot):
    # OpenID standard fields
    sub: str = ff.id_(validators=[ff.HasLength(36)])
    name: str = ff.optional()
    given_name: str = ff.optional()
    family_name: str = ff.optional()
    middle_name: str = ff.optional()
    nickname: str = ff.optional()
    preferred_username: str = ff.optional(index=True)
    profile: str = ff.optional()
    picture: str = ff.optional()
    website: str = ff.optional()
    email: str = ff.optional(validators=[ff.IsValidEmail()], index=True, unique=True)
    email_verified: bool = ff.optional(default=False)
    gender: str = ff.optional(validators=[ff.IsOneOf(('Male', 'Female'))])
    birthdate: date = ff.optional()
    zoneinfo: str = ff.optional()
    locale: str = ff.optional()
    phone_number: str = ff.optional()
    phone_number_verified: bool = ff.optional(default=False)
    address: Address = ff.optional()
    updated_at: datetime = ff.now()

    # Custom fields
    created_at: datetime = ff.now(internal=True)
    deleted_at: datetime = ff.optional(internal=True)
    password_hash: str = ff.optional(length=32, internal=True)
    salt: str = ff.optional(internal=True)
    roles: List[Role] = ff.list_(internal=True)
    tenant: Tenant = ff.optional(index=True, internal=True)
    tenant_id: str = ff.optional(index=True, internal=True)

    @classmethod
    def create(cls, **kwargs):
        if 'email' in kwargs:
            kwargs['email'] = str(kwargs['email']).lower()
        if kwargs.get('password'):
            kwargs['salt'] = bcrypt.gensalt().decode()
            kwargs['password_hash'] = User._hash_password(kwargs['password'], kwargs['salt'])
        try:
            kwargs['tenant_id'] = kwargs['tenant'].id
        except KeyError:
            raise ff.MissingArgument('Tenant is a required field for Client::create()')

        return cls(**ff.build_argument_list(kwargs, cls))

    @classmethod
    def _hash_password(cls, password: str, salt: str):
        return bcrypt.hashpw(password.encode('utf-8'), salt.encode()).decode('utf-8')

    def change_password(self, new_password: str):
        self.password_hash = self._hash_password(new_password, self.salt)

    def change_email(self, new_email: str):
        self.email = new_email

    def add_role(self, role: Role):
        if isinstance(role, Role):
            self.roles.append(role)
    
    def remove_role(self, role: Role):
        if isinstance(role, Role):
            self.roles.remove(role)

    def correct_password(self, password: str):
        if not password or not self.password_hash:
            return False
        return self.password_hash == User._hash_password(password, self.salt)

    def generate_scrubbed_user(self):
        resp = {
            'sub': self.sub,
            'name': self.name,
            'given_name': self.given_name,
            'family_name': self.family_name,
            'middle_name': self.middle_name,
            'nickname': self.nickname,
            'preferred_username': self.preferred_username,
            'profile': self.profile,
            'picture': self.picture,
            'website': self.website,
            'email': self.email,
            'email_verified': self.email_verified,
            'gender': self.gender,
            'birthdate': self.birthdate,
            'zoneinfo': self.zoneinfo,
            'locale': self.locale,
            'phone_number': self.phone_number,
            'phone_number_verified': self.phone_number_verified,
            'updated_at': self.updated_at,
            'created_at': self.created_at,
        }
        if self.tenant is not None:
            resp['tenant_id'] = self.tenant.id,
        return resp

    def get_scopes(self):
        return [scope.id for role in self.roles for scope in role.scopes]



from __future__ import annotations

import firefly as ff


class AddMethodToHeaders(ff.DomainService):
    _kernel: ff.Kernel = None
    
    def __call__(self, incoming_kwargs: dict, http_method: str = 'POST'):
        incoming_kwargs = self._add_headers_from_kernel(incoming_kwargs)
        try:
            headers = incoming_kwargs['headers']['http_request'].get('headers')
        except KeyError:
            headers = incoming_kwargs['headers']
        headers['method'] = http_method
        return headers

    def _add_headers_from_kernel(self, item: dict):
        if self._kernel.http_request:
            
            headers = self._kernel.http_request.get('headers', {})
            item['headers'] = item.get('headers', {})
            item['headers'].update(headers)
        return item



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
import uuid

import firefly as ff
from firefly_iaaa import domain
from firefly_iaaa.application.api.generic_endpoint import GenericEndpoint


@ff.rest('/iaaa/reset', method='POST', tags=['public'], secured=False)
class ResetPassword(GenericEndpoint):
    _cache: ff.Cache = None
    _subdomain: str = None
    _send_reset_email: domain.SendResetEmail = None

    def __call__(self, **kwargs):
        try:
            username = kwargs['username'].lower()
        except KeyError:
            raise Exception('Missing username/password')

        found_user = self._registry(domain.User).find(lambda x: x.email == username)
        if found_user:
            cache_id = str(uuid.uuid4())
            self._cache.set(cache_id, value={'message': 'reset', 'username': username}, ttl=1800)
            try:
                self._send_reset_email(username, cache_id)
            except Exception as e:
                return False
        return {'message': 'success'}



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
import os
import base64

import firefly as ff

import firefly_di as di
import firefly_iaaa.domain as domain

from firefly_aws import infrastructure as aws_infra
from firefly_iaaa.domain.mock.mock_cache import MockCache
from dotenv import load_dotenv

def secret_key_setter():
    pem = os.environ.get('PEM')
    if os.environ.get('FF_ENVIRONMENT') == 'test':
        pem = os.environ.get('TEST_PEM')
    return str(base64.b64decode(pem), "utf-8")
    

load_dotenv()
class Container(di.Container):
    cache: ff.Cache = MockCache if \
        os.environ.get('FF_ENVIRONMENT') == 'test' else aws_infra.DdbCache
    oauthlib_request_validator: domain.OauthRequestValidators = domain.OauthRequestValidators
    request_validator: domain.OauthProvider = domain.OauthProvider
    message_factory: ff.MessageFactory = ff.MessageFactory
    secret_key: str = lambda x: secret_key_setter()
    subdomain: str = lambda x: os.environ.get('SUBDOMAIN')



from __future__ import annotations

import firefly as ff
import jwt


class DecodeToken(ff.DomainService):
    _secret_key: str = None

    def __call__(self, token: str, audience: str = None, **kwargs):
        if token.lower().startswith('bearer'):
            token = token.split(' ')[-1]
        if not audience:
            decoded = jwt.decode(token, options={"verify_signature": False})
            audience = decoded['aud']
        return jwt.decode(token, self._secret_key, 'HS256', audience=audience)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import os

from botocore.exceptions import ClientError

import firefly as ff
import firefly_iaaa.domain as domain


class OAuthRegister(ff.DomainService):
    _oauth_login: domain.OAuthLogin = None
    _registry: ff.Registry = None
    _make_user: domain.MakeClientUserEntities = None
    _context_map: ff.ContextMap = None
    _context: str = None
    _user_created_event = os.environ.get('USER_CREATED_EVENT') or 'iaaa.UserCreated'

    def __call__(self, passed_in_kwargs: dict):
        self.info('Registering User')
        username = passed_in_kwargs['username']

        module_name = self._context_map.\
            get_context(self._context).\
            config.\
            get('domain_module', '{}.domain')
        module = importlib.import_module(module_name.format(self._context))
        user_entity = module.__dict__.get('User')

        try:
            found_user = self._registry(user_entity).find(lambda x: x.email == username)
        except ClientError as e:
            if e.response['Error']['Code'] == 'BadRequestException':
                if 'syntax error at or near ")"' in str(e):
                    return {'message': 'error', 'error': 'User already exists'}
            raise e
        if found_user:
            return {'message': 'error','error': 'User already exists'}

        passed_in_kwargs.update({
            'tenant_name': f'user_tenant_{username}',
            'grant_type': 'password',
            'scopes': []
        })

        user = self._make_user(**passed_in_kwargs)
        self.dispatch(self._user_created_event, data=user.to_dict())

        return True



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
from typing import List, Union

import firefly as ff

from firefly_iaaa.domain.service.decode_token import DecodeToken


class GenericEndpoint(ff.ApplicationService):
    _registry: ff.Registry = None
    _decode_token: DecodeToken = None

    def __call__(self, **kwargs):
        pass

    @staticmethod
    def _make_response(data: Union[dict, ff.Envelope] = None, headers: dict = None, forwarding_address: str = None, cookies: List[dict] = None):
        if isinstance(data, ff.Envelope):
            message = data
        else:
            message = {'message': 'success'}
            if data:
                message['data'] = data
            message = ff.Envelope.wrap(message)
        if headers:
            message = message.set_raw_request(headers)
        if forwarding_address:
            message = message.add_forwarding_address(forwarding_address)
        # if cookies:
        #     message = message.set_cookies(cookies)
        return message

    @staticmethod
    def _make_error_response(error: Union[str, dict]):
        resp = {'message': 'error'}
        resp['error'] = error if isinstance(error, str) else error['error']
        return ff.Envelope.wrap(resp)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
import os
import uuid

import firefly as ff
from firefly_iaaa.application.api.generic_oauth_iam_endpoint import GenericOauthIamEndpoint
import firefly_iaaa.domain as domain


@ff.rest('/iaaa/make-client', method='POST', secured=True)
class MakeClient(GenericOauthIamEndpoint):
    _registry: ff.Registry = None

    def __call__(self, **kwargs):
        roles = []
        for role in kwargs['roles']:
            r = self._registry(domain.Role).find(lambda x: x.name == role)
            roles.append(r)
        kwargs['roles'] = roles

        tenant = domain.Tenant(
            name=kwargs['tenant_name']
        )
        kwargs['tenant'] = tenant

        kwargs['client_id'] = str(uuid.uuid4())
        kwargs['name'] = kwargs.get('name', kwargs['tenant_name'])
        kwargs = self._make_params(kwargs)

        if kwargs.get('scopes') is None or len(kwargs['scopes']) == 0:
            kwargs['scopes'] = ['iaaa.default.read']
        scopes = []
        for scope in kwargs['scopes']:
            s = self._registry(domain.Scope).find(scope)
            scopes.append(s)
        kwargs['scopes'] = scopes

        client = domain.Client.create(**kwargs)

        # Append at end to avoid appending before an error during entity creation
        self._registry(domain.Tenant).append(tenant)
        self._registry(domain.Client).append(client)
        return True

    def _make_params(self, kwargs: dict):

        self._validate_base_params(kwargs)
        grant_type = kwargs['grant_type']
        if grant_type == 'authorization_code':
            kwargs = self._add_auth_code_params(kwargs, False)
        elif grant_type == 'authorization_code_w_pkce':
            kwargs = self._add_auth_code_params(kwargs)
            kwargs['grant_type'] = 'authorization_code'
        elif grant_type == 'implicit':
            kwargs = self._add_auth_code_params(kwargs)
            kwargs['allowed_response_types'] = 'token'
        elif grant_type == 'client_credentials':
            kwargs['client_secret'] = str(uuid.uuid4())
        elif grant_type == 'password':
            name = kwargs['name']
            if name.startswith('user_tenant'):
                name = name[11:]
            kwargs['name'] = f'{name}'

        else:
            raise Exception('Invalid grant type')

        return kwargs

    def _validate_base_params(self, kwargs: dict):
        self._check_kwargs_for_fields(['scopes', 'grant_type'], kwargs)

    def _add_auth_code_params(self, kwargs: dict, uses_pkce: bool = True):
        fields = ['default_redirect_uri', 'redirect_uris']
        self._check_kwargs_for_fields(fields, kwargs)
        kwargs.update({
            'allowed_response_types': 'code',
            'uses_pkce': uses_pkce,
        })
        return kwargs

    def _get_consumer_client(self):
        client: domain.Client = self._registry(domain.Client).find(lambda x: x.client_id == os.environ['CONSUMER_CLIENT_ID'])
        tenant = client.tenant
        return client.client_id, tenant

    def _check_kwargs_for_fields(self, fields, kwargs):
        for field in fields:
            if field not in kwargs:
                raise Exception(f'Missing required field: {field}')


#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='firefly-iaaa',
    version='0.1.17',
    author="JD Williams",
    author_email="me@jdwilliams.xyz",
    description="Bounded context for users of your application.",
    long_description=long_description,
    url="https://github.com/firefly-framework/firefly-iaaa",
    packages=setuptools.PEP420PackageFinder.find('src'),
    package_dir={'': 'src'},
    install_requires=[
        'bcrypt>=3.1.7',
        'firefly-framework>=1.2.9',
        'firefly-aws>=1.2.6',
        'oauthlib>=3.1.1',
        'PyJWT[crypto]>=2.1.0',
    ],
    extras_require={
        'AWS Cognito Support': ['firefly-aws>=1.1.0'],
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import firefly as ff
import firefly_iaaa.domain as domain


class GenericOauthDomainMiddleware(ff.DomainService, ff.LoggerAware, ): 
    _kernel: ff.Kernel = None
    _oauth_provider: domain.OauthProvider = None
    _decode_token: domain.DecodeToken = None

    def __call__(self, message: ff.Message, **kwargs):
        pass

    def _retrieve_token_from_http_request(self):
        for k, v in self._kernel.http_request['headers'].items():
            if k.lower() == 'authorization':
                if not v.lower().startswith('bearer'):
                    raise ff.UnauthorizedError()
                return v
        return None

    def _fix_email(self, message):
        if hasattr(message, 'email'):
            message.email = message.email.lower()
        if hasattr(message, 'username'):
            message.username = message.username.lower()
            message.email = message.email if hasattr(message, 'email') else message.username
        return message



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations
from datetime import datetime

from typing import List

import firefly as ff
from firefly_iaaa.domain.entity.client import Client
from firefly_iaaa.domain.entity.user import User
from firefly_iaaa.domain.entity.scope import Scope



class BearerToken(ff.AggregateRoot):
    id_: str = ff.id_()
    client: Client = ff.required(index=True)
    user: User = ff.required(index=True)
    scopes: List[Scope] = ff.list_()
    access_token: str = ff.required(str, length=36, index=True)
    refresh_token: str = ff.required(str, length=36, index=True)
    expires_at: datetime = ff.required()
    refresh_expires_at: datetime = ff.optional()
    created_at: datetime = ff.now()
    activates_at: datetime = ff.optional(default=datetime.utcnow())
    token_type: str = ff.optional(default='Bearer')
    claims: dict = ff.optional()
    is_access_valid: bool = True
    is_valid: bool = True

    def validate_scopes(self, scopes: List[str]):
        if not scopes:
            return False
        for scope in scopes:
            if scope not in self.get_scopes():
                return False
        return True

    def validate_access_token(self, access_token: str, client: Client):
        return self.access_token == access_token and self.is_access_valid and self._check_active() and self.client.client_id == client.client_id

    def validate_refresh_token(self, refresh_token: str, client: Client):
        return self.refresh_token == refresh_token and self.is_valid and self._check_refresh_active() and self.client == client

    def validate(self, scopes: List[str]):
        return self.token_type.lower() == 'bearer' and self.is_valid and self.validate_scopes(scopes) and self.is_access_valid and self._check_active()

    def invalidate_access_token(self):
        self.is_access_valid = False

    def invalidate(self):
        self.invalidate_access_token()
        self.is_valid = False

    def _has_refresh_expired(self):
        return self.refresh_expires_at < datetime.utcnow() if self.refresh_expires_at is not None else False

    def _has_expired(self):
        return self.expires_at < datetime.utcnow() if self.expires_at is not None else False

    def _has_activated(self):
        return self.activates_at < datetime.utcnow()

    def _check_active(self):
        return self._has_activated() and not self._has_expired()

    def _check_refresh_active(self):
        return self._has_activated() and not self._has_refresh_expired()

    def get_scopes(self):
        return [scope.id for scope in self.scopes]



from __future__ import annotations

import firefly as ff

from firefly_iaaa import domain
from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint


@ff.rest(
    '/iaaa/token', method='POST', tags=['public'], secured=False
)
class OauthTokenCreationService(GenericOauthEndpoint):
    _create_token: domain.CreateToken = None

    def __call__(self, **kwargs):
        kwargs = self._fix_email(kwargs)
        headers, body = self._create_token(kwargs)
        return self._make_response(body, headers)



from __future__ import annotations

import firefly as ff
import json
from firefly_iaaa.application.api.generic_oauth_endpoint import GenericOauthEndpoint


@ff.rest(
    '/iaaa/introspect', method='POST', tags=['public'], secured=True
)
class OauthTokenIntrospectionService(GenericOauthEndpoint):

    def __call__(self, **kwargs):
        kwargs = self._fix_email(kwargs)
        message = self._make_message(kwargs)

        headers, body, status =  self._oauth_provider.create_introspect_response(message)
        
        return self._make_response(json.loads(body), headers)

    def _make_message(self, incoming_kwargs: dict):
        headers = self._add_method_to_headers(incoming_kwargs)
        message_body = {
            'headers': headers,
            'client_id': self._get_client_id(incoming_kwargs.get('client_id')),
            'state': incoming_kwargs.get('state'),
            'token': incoming_kwargs.get('token')
        }

        if incoming_kwargs.get('username'):
            message_body['username'] = incoming_kwargs.get('username') 
        if incoming_kwargs.get('password'):
            message_body['password'] = incoming_kwargs.get('password') 
        if incoming_kwargs.get('client_secret'):
            message_body['client_secret'] = incoming_kwargs.get('client_secret') 
        if not message_body['token']:
            if incoming_kwargs.get('access_token'):
                message_body['token'] = incoming_kwargs.get('access_token') 
        if not message_body['token']:
            if incoming_kwargs.get('refresh_token'):
                message_body['token'] = incoming_kwargs.get('refresh_token')
        if incoming_kwargs.get('token_type_hint'):
            message_body['token_type_hint'] = incoming_kwargs.get('token_type_hint')

        return self._message_factory.query(
            name='OauthIntrospectTokenMessage',
            data=message_body
        )
